# Copyright 2020-2024 Intel Corporation
# SPDX-License-Identifier: Apache-2.0


"""openfl.experimental.workflow.component.collaborator package."""

# FIXME: Too much recursion.
from openfl.experimental.workflow.component.collaborator.collaborator import Collaborator
