# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: openfl/experimental/workflow/protocols/aggregator.proto
# Protobuf Python Version: 5.26.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from openfl.protocols import base_pb2 as openfl_dot_protocols_dot_base__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n7openfl/experimental/workflow/protocols/aggregator.proto\x12\x1copenfl.experiment.aggregator\x1a\x1bopenfl/protocols/base.proto\"o\n\rMessageHeader\x12\x0e\n\x06sender\x18\x01 \x01(\t\x12\x10\n\x08receiver\x18\x02 \x01(\t\x12\x17\n\x0f\x66\x65\x64\x65ration_uuid\x18\x03 \x01(\t\x12#\n\x1bsingle_col_cert_common_name\x18\x04 \x01(\t\"\xae\x01\n\x12TaskResultsRequest\x12;\n\x06header\x18\x01 \x01(\x0b\x32+.openfl.experiment.aggregator.MessageHeader\x12\x13\n\x0b\x63ollab_name\x18\x02 \x01(\t\x12\x14\n\x0cround_number\x18\x03 \x01(\x05\x12\x11\n\tnext_step\x18\x04 \x01(\t\x12\x1d\n\x15\x65xecution_environment\x18\x05 \x01(\x0c\"R\n\x13TaskResultsResponse\x12;\n\x06header\x18\x01 \x01(\x0b\x32+.openfl.experiment.aggregator.MessageHeader\"N\n\x0fGetTasksRequest\x12;\n\x06header\x18\x01 \x01(\x0b\x32+.openfl.experiment.aggregator.MessageHeader\"\xbd\x01\n\x10GetTasksResponse\x12;\n\x06header\x18\x01 \x01(\x0b\x32+.openfl.experiment.aggregator.MessageHeader\x12\x14\n\x0cround_number\x18\x02 \x01(\x05\x12\x15\n\rfunction_name\x18\x03 \x01(\t\x12\x1d\n\x15\x65xecution_environment\x18\x04 \x01(\x0c\x12\x12\n\nsleep_time\x18\x05 \x01(\x05\x12\x0c\n\x04quit\x18\x06 \x01(\x08\"\x98\x01\n\x11\x43heckpointRequest\x12;\n\x06header\x18\x01 \x01(\x0b\x32+.openfl.experiment.aggregator.MessageHeader\x12\x1d\n\x15\x65xecution_environment\x18\x02 \x01(\x0c\x12\x10\n\x08\x66unction\x18\x03 \x01(\x0c\x12\x15\n\rstream_buffer\x18\x04 \x01(\x0c\"Q\n\x12\x43heckpointResponse\x12;\n\x06header\x18\x01 \x01(\x0b\x32+.openfl.experiment.aggregator.MessageHeader2\x84\x02\n\nAggregator\x12U\n\x0fSendTaskResults\x12\x0b.DataStream\x1a\x31.openfl.experiment.aggregator.TaskResultsResponse\"\x00(\x01\x12J\n\x08GetTasks\x12-.openfl.experiment.aggregator.GetTasksRequest\x1a\x0b.DataStream\"\x00\x30\x01\x12S\n\x0e\x43\x61llCheckpoint\x12\x0b.DataStream\x1a\x30.openfl.experiment.aggregator.CheckpointResponse\"\x00(\x01\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'openfl.experimental.workflow.protocols.aggregator_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_MESSAGEHEADER']._serialized_start=118
  _globals['_MESSAGEHEADER']._serialized_end=229
  _globals['_TASKRESULTSREQUEST']._serialized_start=232
  _globals['_TASKRESULTSREQUEST']._serialized_end=406
  _globals['_TASKRESULTSRESPONSE']._serialized_start=408
  _globals['_TASKRESULTSRESPONSE']._serialized_end=490
  _globals['_GETTASKSREQUEST']._serialized_start=492
  _globals['_GETTASKSREQUEST']._serialized_end=570
  _globals['_GETTASKSRESPONSE']._serialized_start=573
  _globals['_GETTASKSRESPONSE']._serialized_end=762
  _globals['_CHECKPOINTREQUEST']._serialized_start=765
  _globals['_CHECKPOINTREQUEST']._serialized_end=917
  _globals['_CHECKPOINTRESPONSE']._serialized_start=919
  _globals['_CHECKPOINTRESPONSE']._serialized_end=1000
  _globals['_AGGREGATOR']._serialized_start=1003
  _globals['_AGGREGATOR']._serialized_end=1263
# @@protoc_insertion_point(module_scope)
